using System;
using Dakota.Data;
using Dakota.Renderables;
using Dakota.Utils;
using Nez;
using Random = Nez.Random;
using Vector2 = Microsoft.Xna.Framework.Vector2;

namespace Dakota.Space
{
    public enum AsteroidSize
    {
        SMALL,
        MEDIUM,
        LARGE,
        HUGE
    }

    public class Asteroid : Component, ITriggerListener
    {
        public AsteroidSize Size;

        public Asteroid(AsteroidSize size = AsteroidSize.LARGE)
        {
            Size = size;
        }

        void ITriggerListener.OnTriggerEnter(Collider other, Collider self)
        {
            if (other.Entity.Name == "bullet" && !other.Entity.IsDestroyed)
            {
                var scene = (SpaceScene) Entity.Scene;
                scene.HitAsteroid(self.Entity, other.Entity.Position);
                other.Entity.Destroy();
            }

            // need to only process A-A collisions once, so use ID comparison
            if (other.Entity.Name == "asteroid" && other.Entity.Id > self.Entity.Id)
            {
                other.CollidesWith(self, out var collisionResult);
                self.Entity.Position += collisionResult.MinimumTranslationVector;
                var p1 = self.Entity.Position;
                var v1 = self.Entity.GetComponent<SpaceObject>().Velocity;
                var p2 = other.Entity.Position;
                var v2 = other.Entity.GetComponent<SpaceObject>().Velocity;
                var diff1 = p1 - p2;
                var diff2 = p2 - p1;
                var newV1 = v1 - Vector2.Dot(v1 - v2, diff1) / diff1.LengthSquared() * diff1;
                var newV2 = v2 - Vector2.Dot(v2 - v1, diff2) / diff2.LengthSquared() * diff2;
                self.Entity.GetComponent<SpaceObject>().Velocity = newV1;
                other.Entity.GetComponent<SpaceObject>().Velocity = newV2;
            }
        }

        void ITriggerListener.OnTriggerExit(Collider other, Collider self)
        {
        }

        public override void OnAddedToEntity()
        {
            var numPoints = Random.NextInt(3) + 7;
            var polygon = new Vector2[numPoints];
            for (var i = 0; i < numPoints; i++)
            {
                var angle = (float) ((float) i / numPoints * 2 * Math.PI);
                var radius = Radius() + Random.MinusOneToOne() * Radius() * 0.2f;
                polygon[i] = radius * DakotaUtils.VecFromAngle(angle);
            }

            // need to clone polygon here so original is intact for drawing rotations
            Entity.AddComponent(new PolygonCollider((Vector2[]) polygon.Clone())).IsTrigger = true;
            Entity.AddComponent(new VectorSprite(polygon, GameConstants.AsteroidColor));
        }

        public float Radius()
        {
            return Size switch
            {
                AsteroidSize.HUGE => 128,
                AsteroidSize.LARGE => 72,
                AsteroidSize.MEDIUM => 48,
                AsteroidSize.SMALL => 24,
                _ => 256
            };
        }
    }
}